/*
 * Copyright (c) 2014, 2017, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.svm.core.heap;

import org.graalvm.compiler.word.ObjectAccess;
import org.graalvm.word.Pointer;
import org.graalvm.word.WordFactory;

import com.oracle.svm.core.UnsafeAccess;
import com.oracle.svm.core.annotate.Alias;
import com.oracle.svm.core.annotate.ExcludeFromReferenceMap;
import com.oracle.svm.core.annotate.RecomputeFieldValue;
import com.oracle.svm.core.annotate.RecomputeFieldValue.Kind;
import com.oracle.svm.core.annotate.TargetClass;
import com.oracle.svm.core.annotate.Uninterruptible;
import com.oracle.svm.core.util.VMError;

/**
 * This class is the plumbing under java.lang.ref.Reference. Instances of this class are discovered
 * by the collector and put on a list. The list can then be inspected to implement
 * java.lang.ref.Reference-like operations.
 *
 * Instances of DiscoverableReference are treated specially by the collector.
 * <ul>
 * <li>The "referent" field will *not* be followed during live object discovery. It is a Pointer,
 * not an Object.</li>
 * <li>During live object discovery, each DiscoverableReference will be added to the "discovered"
 * list.</li>
 * <li>If the object referenced by the referent field is not live at the end of a collection, the
 * "referent" field will be set to null.</li>
 * <li>If the object referenced by the referent field is live at the end of a collection, the
 * referent field is updated if the object moved.</li>
 * <li>If the object referenced by the referent field is live at the end of a collection, the
 * DiscoverableReference will be removed from the discovered list.</li>
 * </ul>
 * After a collection, anyone who is interested can run down the discovered list and do something
 * with each DiscoverableReference.
 *
 * On top of this, I think you could build PhantomReference and WeakReference classes by adding a
 * getReferent() method that returned the referent as an Object. See {@linkplain FeebleReference}
 * and {@linkplain FeebleReferenceList}. Reference classes that subsequently promote their referent
 * (SoftReference and FinalReference) would be more complicated.
 *
 * All the state that is needed to build lists, etc., is allocated in the DiscoverableReference
 * instances, because space can not be allocated for lists during a collection.
 */
public class DiscoverableReference {

    /*
     * List access methods.
     */

    /** Push a DiscoverableReference onto the list. */
    public DiscoverableReference prependToDiscoveredReference(DiscoverableReference newNext) {
        setNextDiscoverableReference(newNext, true);
        return this;
    }

    /*
     * Instance access methods.
     */

    /** Read access to the referent, as an Object. */
    Object getReferentObject() {
        return rawReferent;
    }

    /** Clear the referent. */
    public void clear() {
        rawReferent = null;
    }

    /**
     * Read access to the referent, as a Pointer. This is the low-level access for the garbage
     * collector, so no barriers are used.
     */
    public Pointer getReferentPointer() {
        return ObjectAccess.readWord(this, WordFactory.signed(RAW_REFERENT_OFFSET));
    }

    /**
     * Write access to the referent, as a Pointer. This is the low-level access for the garbage
     * collector, so no barriers are used.
     */
    public void setReferentPointer(Pointer value) {
        ObjectAccess.writeWord(this, WordFactory.signed(RAW_REFERENT_OFFSET), value);
    }

    /** Read access to the next field. */
    public DiscoverableReference getNextDiscoverableReference() {
        return next;
    }

    /** Write access to the next field. */
    private void setNextDiscoverableReference(DiscoverableReference newNext, boolean newIsDiscovered) {
        next = newNext;
        isDiscovered = newIsDiscovered;
    }

    public boolean getIsDiscovered() {
        return isDiscovered;
    }

    public void clean() {
        setNextDiscoverableReference(null, false);
    }

    /**
     * Constructor for sub-classes.
     *
     * Only sub-types of this base type should be constructed.
     *
     * @param referent The Object to be tracked by this DiscoverableReference.
     */
    protected DiscoverableReference(Object referent) {
        initialize(referent);
    }

    @Uninterruptible(reason = "The initialization of the fields must be atomic with respect to collection.")
    private void initialize(Object referent) {
        this.rawReferent = referent;
        this.next = null;
        this.isDiscovered = false;
    }

    /*
     * Instance state.
     *
     * Private rather than protected because users make subclasses of this class.
     */

    /**
     * The referent. The declared type must be {@link Object}, so that the static analysis can track
     * reads and writes. However, the field must not be in the regular reference map since we do all
     * the garbage collection support manually. The garbage collector performs Pointer-level access
     * to the field. This is fine from the point of view of the static analysis, since field stores
     * by the garbage collector do not change the type of the referent.
     */
    @ExcludeFromReferenceMap("Field is manually processed by the garbage collector.") //
    private Object rawReferent;

    /**
     * The offset of the field {@link #rawReferent} for Pointer-level access to the field.
     */
    private static final long RAW_REFERENT_OFFSET;

    static {
        try {
            /*
             * Step 1 of offset computation: this computes the offset on the hosting VM during
             * native image generation. It is not yet the correct offset at run time.
             */
            RAW_REFERENT_OFFSET = UnsafeAccess.UNSAFE.objectFieldOffset(DiscoverableReference.class.getDeclaredField("rawReferent"));
        } catch (NoSuchFieldException ex) {
            throw VMError.shouldNotReachHere(ex);
        }
    }

    @TargetClass(DiscoverableReference.class)
    static final class Target_com_oracle_svm_core_heap_DiscoverableReference {
        // Checkstyle: stop
        /*
         * Step 2 of offset computation: rewrite the field to the offset used at run time. Using the
         *
         * @RecomputeFieldValue is easier than using a Feature class to get the run-time field
         * offset.
         */
        @Alias @RecomputeFieldValue(kind = Kind.TranslateFieldOffset, declClass = DiscoverableReference.class) //
        private static long RAW_REFERENT_OFFSET;
        // Checkstyle: resume
    }

    /**
     * Is this DiscoverableReference on a list?
     * <p>
     * DiscoverableReference does not use the self-link secret of the ancients that FeebleReference
     * uses, because the discovery happens during blackening, so the DiscoverableReference has been
     * promoted, but the next field has not yet been updated, so this == next fails.
     */
    private boolean isDiscovered;
    /** The next element in whichever list of DiscoverableReferences. */
    private DiscoverableReference next;

    /** For testing and debugging. */
    public static final class TestingBackDoor {

        private TestingBackDoor() {
            /* No instances. */
        }

        public static Pointer getReferentPointer(DiscoverableReference that) {
            return that.getReferentPointer();
        }

        public static DiscoverableReference getNextDiscoverableReference(DiscoverableReference that) {
            return that.getNextDiscoverableReference();
        }
    }
}
